LABELS = {
    "storage.googleapis.com": "Google API Storage",
    "v.vrv.co": "VRV",
    "pl.crunchyroll.com": "Crunchyroll",
}

SOURCE_REPOSITORY = "justfoolingaround", "animdl"
VERSION_FILE_PATH = "master", "animdl/core/__version__.py"



import threading

import regex
import yarl

from ...helper import uwu

CONTENT_ID_REGEX = regex.compile(r"embed-6/([^?#&/.]+)")


parent_thread = threading.current_thread()

sid_holder = {
    "sid": None,
}

active_event = threading.Event()


def extract(session, url, **opts):

    from .polling import ws_stimulation

    if sid_holder["sid"] is None and not active_event.is_set():
        active_event.set()

        threading.Thread(
            target=ws_stimulation,
            kwargs={
                "session": session,
                "close_event": active_event,
                "sid_holder": sid_holder,
                "parent_thread": parent_thread,
            },
        ).start()

    while sid_holder["sid"] is None:
        pass

    sid = sid_holder["sid"]

    content_id = CONTENT_ID_REGEX.search(url).group(1)
    recaptcha_response = uwu.bypass_recaptcha(session, url, opts["headers"])

    ajax_response = session.get(
        "https://{}/ajax/embed-6/getSources".format(yarl.URL(url).host),
        params={
            "id": content_id,
            "_token": recaptcha_response.get("token", ""),
            "_number": recaptcha_response.get("number", ""),
        },
    )

    sources = ajax_response.json()

    subtitles = [
        _.get("file") for _ in sources.get("tracks") if _.get("kind") == "captions"
    ]

    def yielder():
        for _ in sources.get("sources"):
            yield {
                "stream_url": _.get("file"),
                "subtitle": subtitles,
                "headers": {"SID": sid},
            }

        for _ in sources.get("sourcesBackup"):
            yield {
                "stream_url": _.get("file"),
                "subtitle": subtitles,
                "headers": {"SID": sid},
            }

    return list(yielder())



import functools
import time

import pypresence

ANIMDL_APPLICATION_ID = 897850955373105222

try:
    presence_client = pypresence.Presence(ANIMDL_APPLICATION_ID)
except (pypresence.DiscordNotFound, ConnectionError):
    raise RuntimeError("Could not connect to Discord.")

presence_client.connect()

presence_client.update(
    state="Queuing up the streams",
)

KITSU_ENDPOINT = "https://kitsu.io/api/"


@functools.lru_cache()
def get_anime(session, anime_name):
    return session.get(
        KITSU_ENDPOINT + "edge/anime", params={"filter[text]": anime_name}
    ).json()["data"]


@functools.lru_cache()
def get_episodes(session, anime_id, offset):

    params = {
        "filter[media_id]": anime_id,
        "sort": "number",
        "filter[mediaType]": "Anime",
        "page[limit]": 10,
    }

    if offset is not None and offset > 1:
        params["page[offset]"] = offset

    return session.get(KITSU_ENDPOINT + "edge/episodes", params=params).json()["data"]


def set_streaming_episode(session, anime_name, episode):

    state = anime_name

    if episode:
        state += ": Episode {}".format(episode)

    content_list = get_anime(session, anime_name)

    if not content_list:
        return presence_client.update(
            state=state,
            large_image="mascot",
            start=int(time.time()),
        )

    anime = content_list[0]

    anime_attributes = anime["attributes"]

    image = anime_attributes["posterImage"]["original"]
    count = anime_attributes["episodeCount"]

    if count is None or (episode > count):
        return presence_client.update(
            state=state,
            large_image=image,
            small_image="mascot",
            large_text="https://github.com/justfoolingaround/animdl",
            start=int(time.time()),
        )

    around = episode - (episode % 10)

    current = get_episodes(session, anime["id"], around)[episode % 10 - 1]

    episode_thumbnail = (current.get("attributes", {}).get("thumbnail", {}) or {}).get(
        "original", "mascot"
    )

    if episode_thumbnail == "mascot":
        image, episode_thumbnail = episode_thumbnail, image

    return presence_client.update(
        state=state + ("/{}".format(count)) if count else "~",
        large_image=episode_thumbnail,
        small_image=image,
        details=current.get("attributes", {}).get("canonicalTitle"),
        start=int(time.time()),
    )



import click
from anchor.strings.parsing.ranges import iter_conditions

from .banner import banner_gift_wrapper
from .logger import setup_loggers
from .player import PLAYER_MAPPING

accept_all = lambda *args, **kwargs: True


class RangeParameter(click.ParamType):

    name = "rangetype"

    def convert(self, value, param, ctx):

        if value is None:
            return None

        conditions = list(iter_conditions(value))

        if not conditions:
            return lambda *args, **kwargs: True

        return lambda *args, **kwargs: any(
            condition(*args, **kwargs) for condition in conditions
        )


def options_stack(options=[]):
    def __inner__(func):
        for option in options:
            func = option(func)

        return func

    return __inner__


def logging_options(
    log_file_args=("--log-file",),
    log_level_args=("--log-level", "-ll"),
):

    return options_stack(
        [
            click.option(
                *log_file_args,
                help="Set a log file to log everything to.",
                required=False,
            ),
            click.option(
                *log_level_args,
                help="Set the integer log level.",
                type=int,
                default=20,
            ),
        ]
    )


def content_fetch_options(
    query_argument_name="query",
    range_args=("-r", "--range"),
    *,
    include_quality_options=True,
    quality_args=("-q", "--quality"),
    include_special_options=True,
    special_args=("-s", "--special"),
    default_quality_string=None,
):

    options = [
        click.argument(
            query_argument_name,
            required=True,
        ),
        click.option(
            *range_args,
            help="Select ranges of anime.",
            required=False,
            default=":",
            type=RangeParameter(),
        ),
    ]

    if include_special_options:
        options.append(
            click.option(
                *special_args,
                help="Special range selection.",
                required=False,
                default="",
                type=str,
            ),
        )

    if include_quality_options:
        options.append(
            click.option(
                *quality_args,
                help="Use quality strings.",
                required=False,
                default=default_quality_string,
            )
        )

    return options_stack(options)


def automatic_selection_options(
    index_args=("--index",),
):

    return options_stack(
        [
            click.option(
                *index_args,
                required=False,
                default=None,
                show_default=False,
                type=int,
                help="Index for the auto flag.",
            ),
        ]
    )


def download_options(
    download_dir_args=("-d", "--download-dir"),
    idm_args=("--idm",),
    default_download_dir=None,
):

    return options_stack(
        [
            click.option(
                *download_dir_args,
                help="Download directory for downloads.",
                required=False,
                default=default_download_dir,
                show_default=False,
                type=click.Path(exists=True, file_okay=False, dir_okay=True),
            ),
            click.option(
                *idm_args,
                is_flag=True,
                default=False,
                help="Download anime using Internet Download Manager",
            ),
        ]
    )


def player_options(
    player_opts_args=("--player-opts",),
    player_args=("-p", "--player"),
    default_player=None,
):

    return options_stack(
        [
            click.option(
                *player_opts_args,
                help="Arguments that are to be passed to the player call.",
                required=False,
            ),
            click.option(
                *player_args,
                help="Select which player to play from.",
                required=False,
                default=default_player,
                show_default=True,
                show_choices=True,
                type=click.Choice(PLAYER_MAPPING.keys(), case_sensitive=False),
            ),
        ]
    )



import functools

import regex

from ...providers.nineanime.decipher import cipher_keyed, encrypt

EMBED_URL_REGEX = regex.compile(r"(.+?/)(?:e(?:mbed)?)/([a-zA-Z0-9]+)")

BASE64_TABLE = "51wJ0FDq/UVCefLopEcmK3ni4WIQztMjZdSYOsbHr9R2h7PvxBGAuglaN8+kXT6y"


CIPHER_KEY_API = "https://raw.githubusercontent.com/justfoolingaround/animdl-provider-benchmarks/master/api/selgen.json"


@functools.lru_cache()
def get_ciper_key(session):
    return session.get(CIPHER_KEY_API).json()["cipher_key"]


def extract(session, url, **opts):
    """
    `CIPHER_KEY_API` created by very shit dev, wonder who that is.

    In any case 9Anime, still wanna play tag?
    """

    match = EMBED_URL_REGEX.search(url)
    host, slug = match.group(1, 2)

    cipher_key = get_ciper_key(session)

    if cipher_key is None:
        return []

    vidstream_info = session.get(
        "{}info/{}".format(
            host,
            encrypt(
                cipher_keyed(cipher_key, encrypt(slug, table=BASE64_TABLE)),
                table=BASE64_TABLE,
            ).replace("/", "_"),
        ),
    )

    if vidstream_info.status_code == 404:
        return []

    return [
        {"stream_url": content.get("file", ""), "headers": {"referer": url}}
        for content in vidstream_info.json()["data"].get("media", {}).get("sources", [])
    ]



import json
import logging
from collections import defaultdict
from functools import partial

import regex
import yarl

from ....config import ALLANIME
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(ALLANIME, extra_regex=r"/anime/([^?&/]+)")

SOURCE_URLS = regex.compile(r'sourceUrl[:=]"(.+?)"')
EPISODES_REGEX = regex.compile(r'\\"availableEpisodesDetail\\":({.+?})')
TITLES_REGEX = regex.compile(r'<span class="mr-1">(.+?);?</span>')

SANITIZER = {
    "\\u002F": "/",
}

EMBED_LIST = (
    ("gogoplay", regex.compile("(streaming|load)\.php\?")),
    ("mp4upload", "https://mp4upload.com/"),
    ("streamsb", "https://streamsb.net/"),
    ("doodstream", "https://dood.to/"),
    ("videobin", "https://videobin.co/"),
    ("okru", "https://ok.ru"),
    ("streamlare", "https://streamlare.com"),
)


def is_embed(url):
    for name, site_url in EMBED_LIST:
        if isinstance(site_url, regex.Pattern):
            if site_url.search(url):
                return name
            continue

        if site_url in url:
            return name
    return None


def iter_episodes(episode_dictionary: dict, anime_page_url: str):
    episodes = defaultdict(list)

    for type_of, episode_numbers in episode_dictionary.items():
        for episode in episode_numbers:
            episodes[int(episode) if episode.isdigit() else 0].append(
                (
                    "Episode {} [{}]".format(episode, type_of.upper()),
                    anime_page_url + "/episodes/{}/{}".format(type_of, episode),
                )
            )

    yield from sorted(episodes.items(), key=lambda x: x[0])


def sanitize(content: "str"):
    for key, item in SANITIZER.items():
        content = content.replace(key, item)
    return content


def extract_content(session, content: "iter_episodes", *, api_endpoint: "str"):
    for title, url in content:
        for source_urls in SOURCE_URLS.finditer(session.get(url).text):
            content_uri = yarl.URL(sanitize(source_urls.group(1)))

            if not content_uri.host:
                link_content = api_endpoint.join(
                    content_uri.with_name(content_uri.name + ".json").with_query(
                        content_uri.query
                    )
                ).human_repr()
                response = session.get(link_content)

                if response.status_code >= 400:
                    continue

                json_response = response.json()

                if not isinstance(json_response, list):
                    json_response = json_response.get("links")

                for link in json_response:
                    stream_uri = yarl.URL(link.get("link"))
                    uri_ref = (
                        stream_uri if stream_uri.host else api_endpoint.join(stream_uri)
                    ).human_repr()
                    yield {
                        "stream_url": uri_ref,
                        "title": title,
                        "headers": {
                            "referer": (
                                api_endpoint.with_path("player").with_query(
                                    {"url": uri_ref}
                                )
                            ).human_repr()
                        },
                    }
            else:
                to_direct = content_uri.human_repr()
                embed = is_embed(to_direct)
                if embed:
                    yield {
                        "stream_url": to_direct,
                        "title": title,
                        "further_extraction": (embed, {}),
                    }
                else:
                    yield {"stream_url": to_direct, "title": title}


def fetcher(session, url: "str", check, match):
    logging.warning("This provider is slow at the cost of high amount of streams.")
    animepage_url = (ALLANIME + "anime/{}").format(match.group(1))

    api_endpoint = yarl.URL(
        session.get(ALLANIME + "getVersion").json().get("episodeIframeHead", "")
    )

    for episode, content in iter_episodes(
        json.loads(
            EPISODES_REGEX.search(session.get(animepage_url).text)
            .group(1)
            .replace('\\"', '"')
        ),
        animepage_url,
    ):
        if check(episode):
            yield partial(
                lambda s, c: list(extract_content(s, c, api_endpoint=api_endpoint)),
                session,
                content,
            ), episode


def metadata_fetcher(session, url: "str", match):
    animepage_url = (ALLANIME + "anime/{}").format(match.group(1))

    return {"titles": TITLES_REGEX.findall(session.get(animepage_url).text)}



import logging
from pathlib import Path

import click

from ...__version__ import __core__
from ...codebase import providers, sanitize_filename
from ...config import (
    AUTO_RETRY,
    DEFAULT_PROVIDER,
    DOWNLOAD_DIRECTORY,
    QUALITY,
    CHECK_FOR_UPDATES,
)
from .. import exit_codes, helpers, http_client


@click.command(name="download", help="Download your favorite anime by query.")
@helpers.decorators.content_fetch_options(
    default_quality_string=QUALITY,
)
@helpers.decorators.download_options(
    default_download_dir=DOWNLOAD_DIRECTORY,
)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    http_client.client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_download(
    query, special, quality, download_dir, idm, index, log_level, **kwargs
):
    r = kwargs.get("range")

    logger = logging.getLogger("downloader")

    anime, provider = helpers.process_query(
        http_client.client, query, logger, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        logger.critical("Searcher returned no anime to stream, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    logger.name = "{}/{}".format(provider, logger.name)

    match, provider_module, _ = providers.get_provider(
        providers.append_protocol(anime.get("anime_url"))
    )

    streams = list(
        provider_module.fetcher(http_client.client, anime.get("anime_url"), r, match)
    )

    if special:
        streams = list(helpers.special.special_parser(streams, special))

    download_directory = Path(download_dir).resolve(strict=True)
    content_name = sanitize_filename(anime["name"])

    content_dir = download_directory / content_name
    content_dir.mkdir(exist_ok=True)

    logger.debug(f"Download directory: {content_dir.as_posix()!r}")
    total = len(streams)

    for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

        content_title = "E{:02d}".format(int(episode_number))
        stream_urls = helpers.ensure_extraction(http_client.client, stream_urls_caller)

        if not stream_urls:
            logger.warning(
                "There were no stream links available at the moment. Ignoring {!r}, retry using a different provider.".format(
                    content_title
                )
            )
            continue

        logger.info(
            "Downloading {!r} [{:02d}/{:02d}, {:02} remaining] ".format(
                content_title, count, total, total - count
            )
        )
        success, reason = helpers.download(
            http_client.client,
            logger,
            content_dir,
            content_title,
            stream_urls,
            quality,
            idm=idm,
            retry_timeout=AUTO_RETRY,
            log_level=log_level,
        )

        if not success:
            logger.warning(
                "Could not download {!r} due to: {}. Please retry with other providers.".format(
                    content_title, reason
                )
            )



from collections import defaultdict
from typing import Callable, DefaultDict, Dict, List, TypeVar

import anitopy
import regex

content_type = TypeVar("content_type")


def construct_site_based_regex(
    site_url: "str", *, extra: "str" = "", extra_regex: "str" = ""
):
    return regex.compile(
        "(?:https?://)?(?:\\S+\\.)*{}".format(
            regex.escape(
                regex.search(r"(?:https?://)?((?:\S+\.)+[^/]+)/?", site_url).group(1)
            )
            + extra
        )
        + extra_regex
    )


def append_protocol(uri: "str", *, protocol: "str" = "https"):
    if regex.search(r"^.+?://", uri):
        return uri
    return "{}://{}".format(protocol.rstrip(":/"), uri.lstrip("/"))


def parse_from_content(
    content: "content_type",
    *,
    name_processor: Callable[[content_type], str] = lambda x: x,
    stream_url_processor: Callable[[content_type], str] = lambda x: x,
    overrides: Dict = {},
    episode_parsed: bool = False
):

    anitopy_result = anitopy.parse(name_processor(content))

    returnee = {"stream_url": stream_url_processor(content)}
    video_res = anitopy_result.get("video_resolution") or ""

    if not episode_parsed:
        returnee.update({"episode": int(anitopy_result.get("episode_number", 0) or 0)})

    if isinstance(video_res, str):
        stripped = video_res.strip("p")
        if stripped.isdigit():
            returnee.update({"quality": int(stripped)})

    returnee.update(overrides)

    return returnee


def group_episodes(
    contents: "List[Dict[str, str]]",
) -> "DefaultDict[int, List[Dict[str, str]]]":
    grouped = defaultdict(list)
    for content in contents:
        grouped[int(content.pop("episode", 0))].append(content)
    return grouped



import json
import logging
import time
from base64 import b64decode, b64encode
from functools import partial

import lxml.html as htmlparser
import regex
import yarl

animixplay_logger = logging.getLogger("provider:animixplay")

ID_MATCHER = regex.compile(r"\?id=([^&]+)")
EMBED_URL_BASE = "https://animixplay.to/api/live"
EMBED_M3U8_MATCHER = regex.compile(r"player\.html[?#](.+?)#")
EMBED_B64_MATCHER = regex.compile(r"#(aHR0[^#]+)")
EMBED_VIDEO_MATCHER = regex.compile(r'iframesrc="(.+?)"')

URL_ALIASES = {
    "bestanimescdn": "omega.kawaiifucdn.xyz/anime3",
    "anicdn.stream": "gogocdn.club",
    "ssload.info": "gogocdn.club",
}


def url_update(url):
    for key, item in URL_ALIASES.items():
        if key in url:
            animixplay_logger.debug(
                "Replacing {!r} to {!r} in {!r}.".format(key, item, url)
            )
            url = url.replace(key, item)
    return url


def extract_from_url(embed_url):
    on_url = EMBED_M3U8_MATCHER.search(embed_url) or EMBED_B64_MATCHER.search(embed_url)

    if not on_url:
        animixplay_logger.debug(
            "Failed to find stream on {!r}, will fallback to API based m3u8.".format(
                embed_url
            )
        )
        return []

    return [{"stream_url": url_update(b64decode(on_url.group(1)).decode())}]


def extract_from_embed(session, embed_url):
    embed_page = session.get(embed_url)

    while embed_page.status_code == 429:
        embed_page = session.get(embed_url)
        time.sleep(2.5)

    if embed_page.status_code == 403:
        return []

    on_site = EMBED_VIDEO_MATCHER.search(embed_page.text)

    if on_site:
        animixplay_logger.debug(
            "Regex matched a video on the site: {!r}, extracting.".format(on_site)
        )
        return extract_from_url(on_site.group(1))

    return extract_from_url(str(embed_page.url))


def get_stream_url(session, data_url):

    component = yarl.URL(data_url)

    if component.host == "www.dailymotion.com":
        return [{"stream_url": data_url, "further_extraction": ("dailymotion", {})}]

    content_id = component.query.get("id")

    if content_id:
        return extract_from_embed(
            session,
            EMBED_URL_BASE
            + b64encode(
                "{}LTXs3GrU8we9O{}".format(
                    content_id,
                    b64encode(content_id.encode()).decode(),
                ).encode()
            ).decode(),
        ) or [
            {
                "stream_url": data_url,
                "further_extraction": ("gogoplay", {}),
            }
        ]

    return extract_from_url(data_url)


def fetcher(session, url, check, match):
    data = json.loads(
        htmlparser.fromstring(session.get(url).content)
        .cssselect("#epslistplace")[0]
        .text
    )

    for value in range(int(data.get("eptotal"))):
        if check(value + 1):
            yield partial(get_stream_url, session, data[str(value)]), value + 1



import functools
from collections import defaultdict

API_ACCESS_TOKEN = "Basic vrvluizpdr2eby+RjSKM17dOLacExxq1HAERdxQDO6+2pHvFHTKKnByPD7b6kZVe1dJXifb6SG5NWMz49ABgJA=="
CLIENT_REFRESH_TOKEN = "vDruXuycqziQocWmI8irARTQa9txXvUKW/LQpGjklYNN1wwJ1dL9KtG19IjPgJVcwSbJ8zUIhlaR4W+IlmY3xiJZ+9nefH4RF4ugWvGWPFq1Q2mcJfxStP/qxfdln2K4UqfOqSjTC/Q8gekarRa9n+fdjUqum6YSV18Coz8gUDQoCTh2ljBwF5mtYZIXGxFn3zC2t8GANTGtmViOLxNNF1axa/m2Rgo8aC3B/tiX+Hg6brP2LtCR/EY0P5IHOB5xClUh1/xqTdD4U4NkYEdfrFB/JZ33+vu/pUHrM/Fa0u7qH6X8w/UtBMkfz/LVOCd8pMgrziOYX+oQTyhFcYncJSaoTsR8V0uEQDu6Z9rkR/xIAe8vq4T8D/PMFOFmgIaDN2UG/Rfh0167Esn9D1PmT8wOyPKeTgmlRuXwwJehn3o="


@functools.lru_cache()
def fetch_api_grant_token(session, api_endpoint):
    return "%(token_type)s %(access_token)s" % (
        session.post(
            api_endpoint + "auth/v1/token",
            data={
                "refresh_token": CLIENT_REFRESH_TOKEN,
                "grant_type": "refresh_token",
                "scope": "all",
            },
            headers={
                "Authorization": API_ACCESS_TOKEN,
            },
        ).json()
    )


def fetch_seasons(session, api_endpoint, media_id, *, predicate=None):
    media = session.get(
        api_endpoint + "content/v1/seasons",
        headers={
            "Authorization": fetch_api_grant_token(session, api_endpoint),
        },
        params={"channel_id": "crunchyroll", "id": media_id, "locale": "en-US"},
    ).json()

    episodes = defaultdict(list)

    for season in media.get("items", []):
        for episode in season.get("episodes", []):
            if predicate is None or predicate(episode):
                episodes[episode["episode_number"]].append(episode)

    return episodes


def get_media_streams(
    session,
    api_endpoint,
    media_id,
    *,
    channel_id="crunchyroll",
    locale="en-US",
    media_type="adaptive_hls"
):
    return session.get(
        api_endpoint + "videos/v1/streams",
        params={
            "channel_id": channel_id,
            "locale": locale,
            "id": media_id,
            "type": media_type,
        },
        headers={
            "Authorization": fetch_api_grant_token(session, api_endpoint),
        },
    ).json()



import logging
import sys
from urllib.parse import quote_plus, urlencode

import httpx

try:
    from .exit_codes import INTERNET_ISSUE
except ImportError:
    from exit_codes import INTERNET_ISSUE

headers = {"User-Agent": "animdl/1.5.84"}

CORS_PROXY = "https://corsproxy.io/"


class AnimeHttpClient(httpx.Client):

    http_logger = logging.getLogger("animdl-http")

    def cf_request(self, method, url, *args, params=None, **kwargs):
        url = CORS_PROXY + "?" + quote_plus(url + "?" + urlencode(params or {}))

        headers = kwargs.pop("headers", {})
        headers.update(referer=url)

        return super().request(method, url, headers=headers, *args, **kwargs)


def httpx_exception():
    hook = sys.excepthook

    def exception_hook(exctype, value, traceback):
        if issubclass(exctype, (httpx.HTTPError)):
            exit(
                AnimeHttpClient.http_logger.critical(
                    "{!r}, this issue originates due to connection issues on your or the servers' side. Retry after troubleshooting connection issues on your side.".format(
                        value
                    )
                )
                or INTERNET_ISSUE
            )

        return hook(exctype, value, traceback)

    sys.excepthook = exception_hook


httpx_exception()

client = AnimeHttpClient(headers=headers, follow_redirects=True, timeout=30.0)



import shutil

import click
import regex
from anchor.strings import iter_contentaware_segments

from .constants import SOURCE_REPOSITORY, VERSION_FILE_PATH

terminal_columns = shutil.get_terminal_size((0, 0)).columns

VERSION_REGEX = regex.compile(r'__core__ = "(.*?)"')


def iter_banner(
    session,
    current_version,
    *,
    check_for_updates=True,
    description="A highly efficient, powerful and fast anime scraper.",
):

    author, repository_name = SOURCE_REPOSITORY

    if terminal_columns:
        processor = lambda _: iter_contentaware_segments(
            str.center(_, terminal_columns), terminal_columns
        )
    else:
        processor = lambda _: iter((_,))

    yield from processor(
        click.style(f"{author}/{repository_name} - v{current_version}", fg="magenta")
    )
    yield from processor(click.style(description, fg="magenta"))

    if check_for_updates:

        branch, version_file = VERSION_FILE_PATH
        upstream_version = VERSION_REGEX.search(
            session.get(
                f"https://raw.githubusercontent.com/{author}/{repository_name}/{branch}/{version_file}"
            ).text
        ).group(1)

        if tuple(upstream_version.split(".")) > tuple(current_version.split(".")):
            yield from processor(
                click.style(f"Update available: {upstream_version}", fg="yellow")
            )


def banner_gift_wrapper(session, current_version, *, check_for_updates=False):
    def wrapper(f):
        def __inner__(*args, log_level, log_file, **kwargs):

            if log_level < 20:
                return f(*args, log_level=log_level, log_file=log_file, **kwargs)

            for _ in iter_banner(
                session,
                current_version,
                check_for_updates=check_for_updates,
            ):
                click.echo(_)

            return f(*args, log_level=log_level, log_file=log_file, **kwargs)

        return __inner__

    return wrapper



from typing import Callable, Iterable, TypeVar

import regex

search_type = TypeVar("search_type")


def search(
    query: "str",
    possibilities: "Iterable[search_type]",
    *,
    processor: "Callable[[search_type], str]" = lambda r: r
):

    pattern = regex.compile(
        r".*?".join(map(regex.escape, query.strip())) + r".*", flags=regex.IGNORECASE
    )

    for search_value in possibilities:
        if pattern.fullmatch(processor(search_value)):
            yield search_value



import json as json_
import logging

import click

from ...__version__ import __core__
from ...codebase.providers import get_provider
from ...config import DEFAULT_PROVIDER, CHECK_FOR_UPDATES
from .. import helpers
from ..http_client import client


@click.command(name="search", help="Search for an anime in the provider.")
@click.argument("query", required=True)
@click.option(
    "-p",
    "--provider",
    help="Provider to search in.",
    default=DEFAULT_PROVIDER,
    type=click.Choice(helpers.provider_searcher_mapping.keys(), case_sensitive=False),
)
@helpers.decorators.logging_options()
@click.option(
    "-j",
    "--json",
    help="Output as json.",
    is_flag=True,
    flag_value=True,
)
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_search(query, json, provider, **kwargs):
    logger = logging.getLogger("searcher")

    match, module, _ = get_provider(query, raise_on_failure=False)

    if module is not None:
        genexp = (
            {
                "name": (
                    module.metadata_fetcher(client, query, match)["titles"] or [None]
                )[0]
                or "",
                "anime_url": query,
            },
        )
    else:
        genexp = helpers.provider_searcher_mapping.get(provider)(client, query)

    for count, search_data in enumerate(genexp, 1):
        if json:
            print(json_.dumps(search_data))
        else:
            logger.info("{0:02d}: {1[name]} {1[anime_url]}".format(count, search_data))



import importlib
import pathlib

from ..helper import append_protocol

EXEMPT = ["__init__.py", "__pycache__"]

if "__path__" in globals():
    __this_path__ = pathlib.Path(__path__[0])
else:
    __this_path__ = pathlib.Path()


def iter_providers(*, exempt=EXEMPT):
    for path in __this_path__.glob("*/"):
        if path.name not in exempt:
            yield importlib.import_module(
                ".{.name}".format(path), package=__name__
            ), path.name


def get_provider(url, *, raise_on_failure=True):
    for provider_module, name in iter_providers():
        match = provider_module.REGEX.match(url)
        if match is not None:
            return match, provider_module, name

    if raise_on_failure:
        raise Exception("Can't find a provider for the url {!r}.")

    return None, None, None


def get_appropriate(session, url, check=lambda *args: True):
    regex_match, provider_module, _ = get_provider(append_protocol(url))
    return provider_module.fetcher(session, url, check, match=regex_match)



from functools import partial

import lxml.html as htmlparser
import regex
import yarl

from ....config import ANIMEOUT
from ...helper import construct_site_based_regex, group_episodes, parse_from_content

REGEX = construct_site_based_regex(ANIMEOUT, extra_regex=r"/([^?&/]+)")

TITLES_REGEX = regex.compile(r'<h1 class="page-title">(.+?)</h1>')


def animeout_stream_url(url: "yarl.URL") -> str:
    return "https://public.animeout.xyz/" + url.with_scheme("").human_repr().lstrip("/")


def fetcher(session, url, check, match):
    animeout_page = session.cf_request("GET", url)
    parsed = htmlparser.fromstring(animeout_page.text)

    downloadables = (
        yarl.URL(_.get("href"))
        for _ in parsed.cssselect('.article-content a[href$="mkv"]')
        if "Download" in _.text_content()
    )

    for episode, content in sorted(
        (
            group_episodes(
                parse_from_content(
                    content,
                    name_processor=lambda c: c.name,
                    stream_url_processor=animeout_stream_url,
                )
                for content in downloadables
            ).items()
        ),
        key=lambda x: x[0],
    ):
        if check(episode):
            yield partial(list, content), episode


def metadata_fetcher(session, url, match):
    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



import logging

import regex
import yarl
from Cryptodome.Cipher import AES

from ...cli.helpers import intelliq

ENCRYPTION_DETECTION_REGEX = regex.compile(r"#EXT-X-KEY:METHOD=([^,]+),")
ENCRYPTION_URL_IV_REGEX = regex.compile(
    r'#EXT-X-KEY:METHOD=(?P<method>[^,]+),URI="(?P<key_uri>.+?)"(?:,IV=(?P<iv>.*))?'
)


STREAM_INFO_REGEX = regex.compile(r"#EXT-X-STREAM-INF(:.*?)?\n+(.+)")
QUALITY_REGEX = regex.compile(r"RESOLUTION=\d+x(\d+)")

INTERNAL_STREAMS_REGEX = regex.compile(r"#EXTINF:.+?\n+(.+)")

HLS_STREAM_EXTENSIONS = ["m3u8", "m3u"]


def get_extension(url):
    _, _, extension = yarl.URL(url).name.rpartition(".")
    return extension


def extract_resolution(quality_string):
    resolution = QUALITY_REGEX.search(quality_string)
    if resolution:
        return resolution.group(1)


def def_iv(initial=1):
    while True:
        yield initial.to_bytes(16, "big")
        initial += 1


def get_decrypter(key, *, iv=b"", default_iv_generator):
    if not iv:
        iv = next(default_iv_generator)
    return AES.new(key, AES.MODE_CBC, iv).decrypt


def unencrypted(m3u8_content):
    st = ENCRYPTION_DETECTION_REGEX.search(m3u8_content)
    return (not bool(st)) or st.group(1) == "NONE"


def extract_encryption(m3u8_content):
    return ENCRYPTION_URL_IV_REGEX.search(m3u8_content).group("key_uri", "iv")


def join_url(parent, child):
    return yarl.URL("{}/{}".format(str(parent).rstrip("/"), str(child).lstrip("/")))


def m3u8_generation(session_init, m3u8_uri):
    m3u8_uri_parent = yarl.URL(m3u8_uri).parent
    response = session_init(m3u8_uri)
    for regex_find in STREAM_INFO_REGEX.finditer(
        response.content.decode("utf-8", errors="ignore")
    ):
        stream_info, content_uri = regex_find.groups()
        url = yarl.URL(content_uri)
        if get_extension(url) in HLS_STREAM_EXTENSIONS:
            if not url.is_absolute():
                content_uri = join_url(m3u8_uri_parent, url)
            yield from m3u8_generation(session_init, content_uri)
        yield {"quality": extract_resolution(stream_info), "stream_url": content_uri}


def resolve_stream(session, logger, q_dicts, quality_string):
    for origin_m3u8 in intelliq.filter_quality(q_dicts, quality_string):
        headers = origin_m3u8.get("headers", {})

        m3u8s = list(
            m3u8_generation(
                lambda s: session.get(str(s), headers=headers),
                origin_m3u8.get("stream_url"),
            )
        )
        for m3u8 in intelliq.filter_quality(m3u8s, quality_string):
            content_response = session.get(str(m3u8.get("stream_url")), headers=headers)
            if content_response.status_code < 400:
                return content_response, origin_m3u8
        content_response = session.get(
            str(origin_m3u8.get("stream_url")), headers=headers
        )
        if content_response.status_code < 400:
            return content_response, origin_m3u8


def hls_yield(session, q_dicts, quality_string, auto_retry=2, *, continuation_index=1):

    logger = logging.getLogger("hls/internal")

    content_response, origin_m3u8 = resolve_stream(
        session, logger, q_dicts, quality_string
    )

    m3u8_data = content_response.content.decode("utf-8", errors="ignore")

    base_uri = yarl.URL(str(content_response.url).rstrip("/") + "/").parent
    encryption_uri, encryption_iv, encryption_data = None, None, b""
    encryption_state = not unencrypted(m3u8_data)

    if encryption_state:
        encryption_uri, encryption_iv = extract_encryption(m3u8_data)
        parsed_uri = yarl.URL(encryption_uri)
        if not parsed_uri.is_absolute():
            parsed_uri = base_uri.join(parsed_uri)
        encryption_key_response = session.get(
            parsed_uri.human_repr(), headers=origin_m3u8.get("headers", {})
        )
        encryption_data = encryption_key_response.content

    internal_streams = INTERNAL_STREAMS_REGEX.findall(m3u8_data)
    total_streams = len(internal_streams)

    stream_iter = iter(
        (yarl.URL(_) for _ in internal_streams[continuation_index - 1 :])
    )
    if encryption_state:
        decryptor = get_decrypter(
            encryption_data,
            iv=encryption_iv or None,
            default_iv_generator=def_iv(continuation_index),
        )

    for current_count, stream in enumerate(stream_iter, continuation_index):

        if not stream.is_absolute():
            stream = base_uri.join(stream)

        ts_response = session.get(
            stream.human_repr(), headers=origin_m3u8.get("headers", {})
        )

        ts_response.raise_for_status()
        ts_data = ts_response.content

        if encryption_state:
            ts_data = decryptor(ts_data)

        yield {
            "bytes": ts_data,
            "total": total_streams,
            "current": current_count,
        }



import functools
import json
import logging
import traceback

import click

from ...codebase import downloader, extractors
from . import decorators, intelliq, special, prompts
from .player import handle_streamer
from .processors import process_query, provider_searcher_mapping

fe_logger = logging.getLogger("further-extraction")


def inherit_stream_meta(parent, streams, *, exempt=["headers", "stream_url"]):
    for stream in streams:
        stream.update({_: __ for _, __ in parent.items() if _ not in exempt})
        yield stream


def further_extraction(session, stream):
    extractor, options = stream.pop("further_extraction", (None, None))
    if extractor is None:
        return [stream]

    for ext_module, ext in extractors.iter_extractors():
        if getattr(ext_module.extract, "disabled", False):
            continue

        if ext == extractor:
            try:
                return functools.reduce(
                    list.__add__,
                    list(
                        further_extraction(session, inherited_stream)
                        for inherited_stream in inherit_stream_meta(
                            stream,
                            ext_module.extract(
                                session, stream.get("stream_url"), **options
                            ),
                        )
                    ),
                    [],
                )
            except Exception as e:
                fe_logger.error(
                    "Extraction from {!r} failed due to: {!r}.".format(ext, e)
                )

    return []


def ensure_extraction(session, stream_uri_caller):
    for stream in stream_uri_caller():
        if "further_extraction" in stream:
            yield from further_extraction(session, stream)
        else:
            yield stream


def download(
    session, logger, content_dir, outfile_name, stream_urls, quality, **kwargs
):

    downloadable_content = intelliq.filter_quality(list(stream_urls), quality)

    if not downloadable_content:
        return False, "No downloadable content found."

    errors = []

    for download_data in iter(downloadable_content):
        if "further_extraction" in download_data:
            try:
                further_status, further_yield = download(
                    session,
                    logger,
                    content_dir,
                    outfile_name,
                    further_extraction(session, download_data),
                    quality,
                    **kwargs
                )
                if further_status:
                    return further_status, further_yield
                continue
            except Exception as e:
                logger.critical(
                    "Could not extract from the stream due to {!r}. falling back to other streams.".format(
                        e
                    )
                )
                continue

        content_url = download_data.get("stream_url")
        content_headers = download_data.get("headers")

        try:
            return True, downloader.handle_download(
                session,
                content_url,
                content_headers,
                content_dir,
                outfile_name,
                preferred_quality=quality,
                subtitles=download_data.get("subtitle", []),
                **kwargs
            )
        except Exception as e:
            logger.critical(
                "Oops, due to {!r}, this stream has been rendered unable to download.".format(
                    e
                )
            )

            errors.append((download_data, e, traceback.format_exc()))

    logger.critical(
        "No downloads were done. Use log level DEBUG or, -ll 0 to get complete tracebacks for the failed downloads."
    )

    for count, (download_data, exception, tb) in enumerate(errors, 1):
        logger.debug(
            json.dumps(
                {
                    "download_count": count,
                    "data": download_data,
                    "exception": repr(exception),
                    "traceback": tb,
                },
                indent=4,
            )
        )

    return False, "All stream links were undownloadable."



"""
THANKS FOR ADDING STREAMSB AND STREAMTAPE DUMMIES
"""

from functools import partial

import lxml.html as htmlparser
import regex

from ....config import ZORO
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(ZORO, extra_regex=r"(/watch)?/[\w-]+-(\d+)")
TITLES_REGEX = regex.compile(r'<h2 class="film-name dynamic-name" .+?>(.+?)</h2>')

XHR_HEADERS = {
    "X-Requested-With": "XMLHttpRequest",
    "Referer": ZORO,
}


def int_or(string, *, default=0):
    if string.isdigit():
        return int(string)
    return default


SERVER_IDS = {4: "rapidvideo", 1: "rapidvideo", 5: "streamsb", 3: "streamtape"}


def extract_episode(session, data_id, title):
    for server in htmlparser.fromstring(
        session.get(
            ZORO + "ajax/v2/episode/servers",
            params={"episodeId": data_id},
            headers=XHR_HEADERS,
        )
        .json()
        .get("html")
    ).cssselect("div.server-item"):
        source_data = session.get(
            ZORO + "ajax/v2/episode/sources",
            params={"id": server.get("data-id")},
            headers=XHR_HEADERS,
        ).json()
        if source_data.get("type") != "iframe":
            yield {
                "stream_url": source_data.get("link"),
                "title": "{} - {}".format(server.get("data-type"), title),
            }
            continue

        cdn = SERVER_IDS.get(source_data.get("server"), "unavailable")

        if cdn in ["streamsb", "streamtape", "unavailable"]:
            continue

        yield {
            "stream_url": source_data.get("link"),
            "further_extraction": (
                cdn,
                {"headers": {"Referer": ZORO}},
            ),
            "title": "{} - {}".format(server.get("data-type", "").upper(), title),
        }


def fetcher(session, url, check, match):
    slug = match.group(2)

    for episode in htmlparser.fromstring(
        session.get(ZORO + "ajax/v2/episode/list/{}".format(slug), headers=XHR_HEADERS)
        .json()
        .get("html")
    ).cssselect("a[title][data-number][data-id]"):
        episode_number = int_or(episode.get("data-number", "") or "")
        if check(episode_number):
            yield partial(
                lambda d_id, t: list(extract_episode(session, d_id, t)),
                d_id=episode.get("data-id"),
                t=episode.get("title"),
            ), episode_number


def metadata_fetcher(session, url, match):
    return {"titles": TITLES_REGEX.findall(session.get(ZORO + match.group(2)).text)}



import base64
import functools

import regex
import yarl

RECAPTCHA_API_JS = "https://www.google.com/recaptcha/api.js"


def bypass_ddos_guard(session, base_uri):
    js_bypass_uri = regex.search(
        r"'(.*?)'", session.get("https://check.ddos-guard.net/check.js").text
    ).group(1)

    session.get(base_uri + js_bypass_uri)


def bypass_recaptcha(session, url, headers):
    domain = (
        base64.b64encode("{0.scheme}://{0.host}:443".format(yarl.URL(url)).encode())
        .decode()
        .strip("=")
        + "."
    )
    initial_page = session.get(url, headers=headers)

    site_key_match = regex.search(r"recaptchaSiteKey = '(.+?)'", initial_page.text)

    if site_key_match is None:
        return {}

    response = {
        "token": get_token_recaptcha(
            session,
            domain,
            site_key_match.group(1),
            "{0.scheme}://{0.host}/".format(yarl.URL(url)),
        )
    }

    number_match = regex.search(r"recaptchaNumber = '(\d+?)'", initial_page.text)

    if number_match:
        response.update(number=number_match.group(1))

    return response


@functools.lru_cache()
def get_token_recaptcha(session, domain, recaptcha_site_key, url):

    recaptcha_out = session.get(
        RECAPTCHA_API_JS,
        params={"render": recaptcha_site_key},
        headers={"referer": url},
    ).text

    v_token_match = regex.search(r"releases/([^/&?#]+)", recaptcha_out)

    if v_token_match is None:
        return

    v_token = v_token_match.group(1)
    anchor_out = session.get(
        "https://www.google.com/recaptcha/api2/anchor",
        params={
            "ar": 1,
            "k": recaptcha_site_key,
            "co": domain,
            "hl": "en",
            "v": v_token,
            "size": "invisible",
            "cb": "kr42069kr",
        },
    ).text

    recaptcha_token_match = regex.search(r'recaptcha-token.+?="(.+?)"', anchor_out)

    if recaptcha_token_match is None:
        return

    recaptcha_token = recaptcha_token_match.group(1)

    token_out = session.post(
        "https://www.google.com/recaptcha/api2/reload",
        params={"k": recaptcha_site_key},
        data={
            "v": v_token,
            "reason": "q",
            "k": recaptcha_site_key,
            "c": recaptcha_token,
            "sa": "",
            "co": domain,
        },
        headers={"referer": "https://www.google.com/recaptcha/api2"},
    ).text

    token_match = regex.search(r'rresp","(.+?)"', token_out)
    if token_match is None:
        return

    return token_match.group(1)



from setuptools import find_packages, setup

from animdl.core.__version__ import __core__

with open("requirements.txt") as requirements_txt:
    requirements = requirements_txt.read().splitlines()


setup(
    name="animdl",
    version=__core__,
    author="kr@justfoolingaround",
    author_email="kr.justfoolingaround@gmail.com",
    description="A highly efficient, fast, powerful and light-weight anime downloader and streamer for your favorite anime.",
    packages=find_packages(),
    url="https://github.com/justfoolingaround/animdl",
    keywords=[
        "stream",
        "cli",
        "anime",
        "download",
    ],
    install_requires=requirements,
    entry_points="""
        [console_scripts]
        animdl=animdl.__main__:__animdl_cli__
    """,
    include_package_data=True,
)
